package org.bdware.sc.boundry.utils;

import com.google.gson.JsonObject;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.doip.codec.digitalObject.DigitalObject;
import org.bdware.doip.codec.digitalObject.Element;
import org.bdware.doip.codec.doipMessage.DoipMessage;
import org.bdware.doip.codec.doipMessage.DoipMessageFactory;
import org.bdware.doip.codec.operations.BasicOperations;
import org.bdware.doip.endpoint.client.ClientConfig;
import org.bdware.doip.endpoint.client.DoipClientImpl;
import org.bdware.doip.endpoint.client.DoipMessageCallback;
import org.bdware.sc.compiler.PermissionStub;
import org.bdware.sc.node.Permission;
import org.bdware.sc.util.JsonUtil;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;

@PermissionStub(permission = Permission.DOIP)
public class DOIPUtil {
    private static final Logger LOGGER = LogManager.getLogger(DOIPUtil.class);
    private final String url;
    private final DoipClientImpl doipClient;

    private DOIPUtil(String url) {
        this.url = url;
        this.doipClient = new DoipClientImpl();
        doipClient.connect(ClientConfig.fromUrl(url));
    }

    public static DOIPUtil createClient(String url) {
        return new DOIPUtil(url);
    }

    public String test(String doi) {
        return "create DOClient And hello " + doi + " World";
    }

    public boolean reconnect() {
        try {
            if (doipClient != null)
                doipClient.reconnect();
            return true;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    private DoipMessage syncGetMessage(DoipMessage message) {
        List<DoipMessage> ret = new ArrayList<>();
        DoipMessage msg = null;
        try {
            doipClient.sendMessage(message, new DoipMessageCallback() {
                @Override
                public void onResult(DoipMessage doipMessage) {
                    ret.add(doipMessage);
                    synchronized (ret) {
                        ret.notify();
                    }
                }
            });
            synchronized (ret) {
                ret.wait(5000);
            }
            if (ret.size() > 0)
                msg = ret.get(0);
            if (msg != null) {
                return msg;
            } else return DoipMessageFactory.createTimeoutResponse(message.requestID);
        } catch (Exception ie) {
            ie.printStackTrace();
            return DoipMessageFactory.createConnectFailedResponse(message.requestID);
        }
    }

    private static String convertDoipMsgToString(DoipMessage ret) {
        try {
            LOGGER.info(new String(ret.body.encodedData));
            DigitalObject respDO = ret.body.getDataAsDigitalObject();
            return respDO.toString();
        } catch (Exception ie) {
            ie.printStackTrace();
            return "send doip message error: " + ie.getMessage();
        }
    }

    public String hello(String repoID) {
        DoipMessage msg = (new DoipMessageFactory.DoipMessageBuilder()).createRequest(repoID, BasicOperations.Hello.getName()).create();
        return convertDoipMsgToString(syncGetMessage(msg));
    }

    public String retrieve(String doi, String args) {
        DoipMessage msg = (new DoipMessageFactory.DoipMessageBuilder()).createRequest(doi, BasicOperations.Retrieve.getName())
                .setBody(args.getBytes(StandardCharsets.UTF_8)).create();
        return convertDoipMsgToString(syncGetMessage(msg));
    }

    public String call(String doi, String action, String args) {
        JsonObject jo = new JsonObject();
        jo.addProperty("action", action);
        jo.addProperty("args", args);
        DoipMessage msg = (new DoipMessageFactory.DoipMessageBuilder())
                .createRequest(doi, BasicOperations.Retrieve.getName())
                .setBody(jo.toString().getBytes(StandardCharsets.UTF_8)).create();
        return convertDoipMsgToString(syncGetMessage(msg));
    }

    public String create(String repoID, String doString) {
        DigitalObject digitalObject = JsonUtil.fromJson(doString, DigitalObject.class);
        for (Element e : digitalObject.elements) {
            if (null != e.dataString) {
                e.setData(e.dataString.getBytes());
            }
            e.dataString = null;
        }
        DoipMessage msg =
                new DoipMessageFactory.DoipMessageBuilder()
                        .createRequest(repoID, BasicOperations.Create.getName())
                        .setBody(digitalObject)
                        .create();
        return convertDoipMsgToString(syncGetMessage(msg));
    }

    public String delete(String doID) {
        DoipMessage msg =
                new DoipMessageFactory.DoipMessageBuilder()
                        .createRequest(doID, BasicOperations.Delete.getName())
                        .create();
        return convertDoipMsgToString(syncGetMessage(msg));
    }

    public String listOperation(String doID) {
        DoipMessage msg =
                new DoipMessageFactory.DoipMessageBuilder()
                        .createRequest(doID, BasicOperations.ListOps.getName())
                        .create();
        return convertDoipMsgToString(syncGetMessage(msg));
    }
}
